//    This file is part of JustMotion.

//    JustMotion is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.

//    JustMotion is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.

#include "detect_loop.h"

DetectLoop::DetectLoop(shared_t *sharedRes, QThread *thr, QObject *parent) : QObject(parent)
{
    pcTimer = 0;
    shared  = sharedRes;

    connect(thr, &QThread::started,  this, &DetectLoop::init);
    connect(thr, &QThread::finished, this, &DetectLoop::deleteLater);

    moveToThread(thr);
}

void DetectLoop::init()
{
    pcTimer = new QTimer(this);
    evTimer = new QTimer(this);
    dTime   = QDateTime::currentDateTimeUtc();

    connect(pcTimer, &QTimer::timeout,  this, &DetectLoop::pcBreak);
    connect(evTimer, &QTimer::timeout,  this, &DetectLoop::evBreak);
    connect(this,    &DetectLoop::loop, this, &DetectLoop::exec);

    pcTimer->start(shared->postSecs * 1000);
    evTimer->start(shared->evMaxSecs * 1000);

    setupBuffDir(shared->buffPath);

    eTimer.start();

    QTimer::singleShot(SEG_SIZE + 4, this, &DetectLoop::exec);
}

void DetectLoop::reset()
{
    eventQue.inQue  = false;
    eventQue.score  = 0;
    eventQue.queAge = 0;

    eventQue.imgPath.clear();
    eventQue.vidList.clear();
    eventQue.timeStamp.clear();
}

void DetectLoop::pcBreak()
{
    prevClips.clear();

    if (!shared->postCmd.isEmpty())
    {
        qInfo() << "---POST_BREAK---";

        if (eventQue.inQue)
        {
            qInfo() << "motion detected, skipping the post command";
        }
        else
        {
            qInfo() << "no motion detected, running post command: " << shared->postCmd;

            auto args = parseArgs(shared->postCmd.toUtf8(), -1);

            if (args.isEmpty())
            {
                qCritical() << "err: did not parse an executable from the post command line.";
            }
            else
            {
                QProcess::execute(args[0], args.mid(1));
            }
        }
    }
}

void DetectLoop::evBreak()
{
    if (eventQue.inQue)
    {
        shared->recList.append(eventQue);

        reset();
    }
}

float DetectLoop::getFloatFromExe(const QByteArray &line)
{
    QString strLine(line);
    QString strNum;

    for (auto chr : strLine)
    {
        if (chr.isDigit() || (chr == '.'))
        {
            strNum.append(chr);
        }
        else
        {
            break;
        }
    }

    auto ok  = false;
    auto res = strNum.toFloat(&ok);

    if (!ok || strNum.isEmpty())
    {
        qCritical() << "err: the image comp command returned unexpected output and couldn't be converted to float";
        qCritical() << "     raw output: " << line;
    }

    return res;
}

QStringList DetectLoop::buildArgs(const QString &prev, const QString &next)
{
    auto args = parseArgs(shared->compCmd.toUtf8(), -1);

    for (auto i = 0; i < args.size(); ++i)
    {
        if (args[i] == PREV_IMG) args[i] = prev;
        if (args[i] == NEXT_IMG) args[i] = next;
    }

    return args;
}

QString DetectLoop::statusLine()
{
    if (eTimer.elapsed() >= 30000)
    {
        emit starving();

        return "STARVED";
    }
    else
    {
        return "OK     ";
    }
}

void DetectLoop::exec()
{
    auto vidPath = shared->buffPath + "/live/" + dTime.toString(DATETIME_FMT) + shared->streamExt;
    auto found   = false;

    for (auto i = 0; i < 100; ++i)
    {
        if (QFileInfo::exists(vidPath))
        {
            found = true; break;
        }
        else
        {
            dTime = dTime.addSecs(1);

            vidPath = shared->buffPath + "/live/" + dTime.toString(DATETIME_FMT) + shared->streamExt;
        }
    }

    if (found)
    {
        auto imgPath = shared->buffPath + "/img/" + QFileInfo(vidPath).baseName() + "-%03d" + ".bmp";

        QProcess    ffmpeg;
        QStringList args;
        QString     imgA;
        QString     imgB;

        args.append("ffmpeg");
        args.append("-hide_banner");
        //args.append("-loglevel");
        //args.append("panic");
        args.append("-y");
        args.append("-i");
        args.append(vidPath);
        args.append("-vf");
        args.append("fps=1/1");
        args.append(imgPath);

        ffmpeg.start(args[0], args.mid(1));
        ffmpeg.waitForFinished();

        for (auto i = 1; i <= SEG_SIZE; ++i)
        {
            auto imgNum  = QString::number(i).rightJustified(3, '0');
            auto imgPath = shared->buffPath + "/img/" + QFileInfo(vidPath).baseName() + "-" + imgNum + ".bmp";

            if (QFileInfo::exists(imgPath))
            {
                if      (imgA.isEmpty()) imgA = imgPath;
                else if (imgB.isEmpty()) imgB = imgPath;
                else
                {
                    eTimer.restart();

                    QProcess extComp;

                    auto compArgs = buildArgs(imgA, imgB);

                    extComp.start(compArgs[0], compArgs.mid(1));
                    extComp.waitForFinished();

                    float score = 0;

                    if (shared->outputType == "stdout")
                    {
                        score = getFloatFromExe(extComp.readAllStandardOutput());
                    }
                    else
                    {
                        score = getFloatFromExe(extComp.readAllStandardError());
                    }

                    qInfo() << compArgs.join(" ") << " --result: " << QString::number(score);

                    if (score >= shared->imgThresh)
                    {
                        qInfo() << "--threshold_meet: " << QString::number(shared->imgThresh);

                        if (eventQue.score < score)
                        {
                            eventQue.imgPath = imgB;
                            eventQue.score   = score;
                        }

                        eventQue.inQue     = true;
                        eventQue.queAge    = 0;
                        eventQue.timeStamp = QDateTime::currentDateTimeUtc().toString(DATETIME_FMT);

                        if (!eventQue.vidList.contains(vidPath))
                        {
                            eventQue.vidList.append(vidPath);
                        }
                    }

                    imgA.clear();
                    imgB.clear();
                }
            }
        }

        emit loop();
    }
    else
    {
        QTimer::singleShot(SEG_SIZE + 4, this, &DetectLoop::exec);

        dTime = QDateTime::currentDateTimeUtc();
    }
}
